
//
//
// Project       : Vorpal-Mail
//
// Source        : server.cc
//
// Creator       : Blane Bramble                    Date          : 23rd October 2000
// Amended       : Blane Bramble                    Date          : 27th March 2001
//
// License       : Copyright Virtua-Web 2000-2001, licensed under GNU General Public License (GPL) v2
//
// SMTP daemon server, utilising SMTP_SESSION class (session.h, session.cc). Provides
// an RFC 821 compliant SMTP server daemon.
//
//
// Changelog:
//
// 5th December 2000
// Additional code to allow Vorpal Mail to function as a replacement for sendmail
// from the command line (various programs seem to call sendmail directly to
// generate emails).
//
// 7th December 2000
// Limit specifying a FROM user for sendmail mode to root. Otherwise use
// the uid and gecos information.
//
// 10th January 2001
// Changed to use C++ streams for output
//
// 4th Feb 2001
// Template classes used for server list.
//
// 2nd March 2001
// Added QUEUE class to queue processing.

#include <iostream>
#include <iomanip>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <getopt.h>
#include <dirent.h>
#include <fcntl.h>
#include <syslog.h>
#include <pwd.h>

#include <netdb.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <netinet/in.h>

#include "mail.h"
#include "array.h"
#include "smtp-client.h"
#include "smtp-server.h"
#include "tlist.h"
#include "string.h"
#include "queue.h"
#include "virtual.h"

#define SMTP_SERVER 1
#define SMTP_CLIENT 2

SMTPSERVER *session = NULL;
SMTPCLIENT *client = NULL;

VIRTUAL_DOMAIN *domain;

TLIST<SMTPSERVER> smtplist;

int run = 0, debug = 0, interval = 0, reloaded = 0;

int control_socket = -1;
long retrytime = 300;				// 5 minute gap before a msg is retried
int cached = 0;
int nodaemon = 0;

static struct option longoptions[] = {
	{ "list", 0, NULL, 'l', },
	{ "help", 0, NULL, 'h', },
	{ "daemon", 0, NULL, 'd', },
	{ "queue", 0, NULL, 'q', },
	{ "dump", 0, NULL, 'm', },
	{ "debug", 2, NULL, 'D', },
	{ "retry", 1, NULL, 'r', },
	{ "nodaemon", 0, NULL, 'n', },
	{ NULL, 0, NULL, 0, },
	};

char *personality[] = {
	"mailq",
	"sendmail",
	"send-mail",
	NULL,
	};

enum	{
	PERS_DEFAULT = -1,
	PERS_MAILQ = 0,
	PERS_SENDMAIL,
	PERS_SEND_MAIL,
	PERS_LAST
	};

void signalman(int sig)
{
int status, pid;
switch(sig)
	{
	case SIGINT:
		syslog(LOG_MAIL + LOG_NOTICE, "SIGINT received. Exiting.\n");
		run = 0;
		break;
	case SIGTERM:
		syslog(LOG_MAIL + LOG_NOTICE, "SIGTERM received. Exiting.\n");
		run = 0;
		break;
	case SIGHUP:
		{
		TWALK<SMTPSERVER> walk;

		delete domain;
		domain = new VIRTUAL_DOMAIN();
		domain->load(CONFIG);

		walk.attach(&smtplist);
		if (walk.start())
			{
			do	{
				walk.get()->setdomain(domain);
				} while(walk.next());
			}
		syslog(LOG_MAIL + LOG_NOTICE, "SIGHUP received. Configuration reloaded.\n");
		}
		break;
	case SIGALRM:
		interval++;
		break;
	}
signal(sig, signalman);
}

int makedaemon(void)
{
if (nodaemon)
	{
	chdir("/");
	umask(0);
	return 1;
	}

switch(fork())
	{
	case -1:				// failed, log msg and exit
		syslog(LOG_MAIL + LOG_EMERG, "** Could not fork process. Cannot run as daemon");
		return 0;
	case 0:					// child, so do more stuff
		if (setsid() == -1)
			{
			syslog(LOG_MAIL + LOG_EMERG, "** Could not setsid. Cannot run as daemon");
			return 0;
			}
		switch(fork())
			{
			case -1:		// failed, log and exit
				syslog(LOG_MAIL + LOG_EMERG, "** Could not fork process (2). Cannot run as daemon");
				return 0;
			case 0:			// grandchild, so do more stuff
				chdir("/");	// change to root
				umask(0);	// haha! he would have got away with it to, if it wasn't for us darn kids
				freopen("/dev/null", "r", stdin);
				freopen("/dev/console", "w", stdout);
				freopen("/dev/console", "w", stderr);
				return 1;	// All worship me, for I am an all powerful Daemon
			default:		// child - so can exit
				return 0;
			}
		break;		
	default:				// parent - so can exit
		return 0;
	}
}

int clientinitialise(char *who)
{
long flags;
struct sockaddr_in sad;

struct hostent *host;

if (control_socket == -1)
	{

	if ((control_socket = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		{
		syslog(LOG_MAIL + LOG_EMERG, "** Could not create socket.\n");
		return 0;
		}


	if ((host = gethostbyname(who)) == NULL)
		{
		if (debug)
			syslog(LOG_MAIL + LOG_INFO, "** Could not get relay by name: %s", who);
		control_socket = -1;
		std::cout << "Can't get relay by name: " << who << endl;
		return 0;
		}

	sad.sin_family = AF_INET;
	//sad.sin_addr.s_addr = ((struct sockaddr_in *)(host->h_addr))->sin_addr.s_addr;
	sad.sin_addr.s_addr = *(unsigned long *)(host->h_addr);
	sad.sin_port = htons(25);

	if (connect(control_socket, (sockaddr *)&sad, sizeof(sad)) < 0)
		{
		syslog(LOG_MAIL + LOG_EMERG, "** Could not connect socket.\n");
		std::cout << "Can't connect to relay: " << who << endl;
		return 0;
		}
	flags = 1;
	ioctl(control_socket, FIONBIO, &flags);
	signal(SIGTERM, signalman);
	signal(SIGHUP, signalman);
	signal(SIGINT, signalman);

	client = new SMTPCLIENT(control_socket, &sad);

	cached = 0;
	}
else
	cached = 1;

client->setdebug(debug);
client->setdomain(domain);

return 1;
}

int initialise(void)
{
long flags;
struct sockaddr_in sad;
struct itimerval timer;

if ((control_socket = socket(AF_INET, SOCK_STREAM, 0)) < 0)	/* create a socket, or fail initialisation */
	return 0;

flags = 1;					/* Set non-blocking I/O mode (flags = 1) */
ioctl(control_socket, FIONBIO, &flags);

sad.sin_family = AF_INET;
sad.sin_addr.s_addr = INADDR_ANY;
sad.sin_port = htons(domain->getport());			/* request a contol socket listening on any addr, port as specified */

if (bind(control_socket, (sockaddr *)&sad, sizeof(sad)) < 0)
	return 0;				/* bind the address to our socket, or fail initialisation */

if (listen(control_socket, 5) != 0)
	return 0;				/* start listening (or fail...) */

signal(SIGTERM, signalman);
signal(SIGHUP, signalman);
signal(SIGINT, signalman);
signal(SIGALRM, signalman);

timer.it_value.tv_sec = 10;
timer.it_interval.tv_sec = 10;

if (setitimer(ITIMER_REAL, &timer, NULL) != 0)
	return 0;

return 1;					/* hey, it all worked! */
}

int clientrun(int cache)
{
struct timeval tv;
int ok = 0;

run = 1;

while(run)
	{
	fd_set fdr, fdw;
	FD_ZERO(&fdr);
	FD_ZERO(&fdw);
	FD_SET(control_socket, &fdr);
	if (client->getstatus() == SMTP_STATUS_DATASEND || client->ispending())
		FD_SET(control_socket, &fdw);
	tv.tv_sec = 10; tv.tv_usec = 0;

//	if (cache)						// no connect msg as re-using socket, so spoof a 250
//		client->process("250 cached connection\r\n");
	if (cache)
		client->sendreset();

	cache = 0;

	if (select(FD_SETSIZE, &fdr, &fdw, NULL, &tv) > 0)
		{
		if (FD_ISSET(control_socket, &fdr))
			{
			char buf[SMTP_MAX_LINE];
			int t;
			if ((t=recv(control_socket, buf, SMTP_MAX_LINE, 0))<1)
				{
				syslog(LOG_MAIL + LOG_INFO, "Closed connection removed.\n");
				run = 0;
				}
			else
				{
/* 0 terminate the input string */
				*(buf + t) = 0;
				client->process(buf);
				switch(client->getstatus())
					{
					case SMTP_STATUS_ABORT:
						if (debug)
							syslog(LOG_MAIL + LOG_INFO, "Transmission failed.\n");
						run = 0;
						break;
					case SMTP_STATUS_COMPLETE:
						if (debug)
							syslog(LOG_MAIL + LOG_INFO, "Transmission complete.\n");
						ok = 1;
						run = 0;
						break;
					}
				}
			}
		if (control_socket != -1 && FD_ISSET(control_socket, &fdw))
			{
			if (client->getstatus() == SMTP_STATUS_DATASEND)
				{
				client->process("\r\n");
				}
			if (client->dosend() < 0)
				{
				run = 0;
				syslog(LOG_MAIL + LOG_INFO, "Send failed.\n");
				}
			}
		}
	}
return ok;
}

int serverrun(void)
{
int n;
struct timeval tv;
TWALK<SMTPSERVER> walk;
int pid, status;

if (!initialise())
	{
	std::cout << "Could not initialise." << endl;
	return 1;
	}

if (!makedaemon())
	return 0;

syslog(LOG_MAIL + LOG_NOTICE, PROGRAM " " VERSION " daemon started");

run = 1;

walk.attach(&smtplist);				// attach to the smtp session list;

while(run)
	{
	fd_set fdr,fdw;
	FD_ZERO(&fdr);
	FD_ZERO(&fdw);
	FD_SET(control_socket,&fdr);

	if (walk.start())
		{
		do	{
			int sock;

			session = walk.get();

			if ((sock = session->getsocket()) != -1)
				{
				FD_SET(sock, &fdr);
				if (session->ispending())
					FD_SET(sock, &fdw);
				if (interval)
					{
					session->pulse();
					interval--;
					}
				}
			} while(walk.next());
		}

	if ((pid = waitpid(-1, &status, WNOHANG)) > 0 && debug)
		syslog(LOG_MAIL + LOG_DEBUG, "Child process %d has exitted", pid);

	tv.tv_sec=10; tv.tv_usec=0;
	if (select(FD_SETSIZE,&fdr,&fdw,NULL,&tv) > 0)
		{
/* New Terminal Connected */
		if (FD_ISSET(control_socket, &fdr))
			{
/* New Socket, Addrlength, Line */
			int ns,al,li;
			struct sockaddr_in sa;
			al = sizeof(sa);
			if ((ns=accept(control_socket, (sockaddr*)&sa,(socklen_t*)&al))>-1)
				{
				long fl=1;
				SMTPSERVER *sess;

				ioctl(ns,FIONBIO,&fl);

				if ((sess = new SMTPSERVER(ns, &sa)) == NULL)
					{
					shutdown(ns, 2);
					close(ns);
					}
				else
					{
					smtplist.add(NULL, sess);
					}
				sess->setdebug(debug);
				sess->setdomain(domain);
				sess->serverstatus(220);
				}
			}

		if (walk.start())
			{
			do	{
				int sock;
				session = walk.get();
				if ((sock = session->getsocket())  != -1 && FD_ISSET(sock, &fdr))
					{
					char buf[SMTP_MAX_LINE];
					int t;
					if ((t=recv(sock, buf, SMTP_MAX_LINE, 0))<1)
						{
						walk.del();
						syslog(LOG_MAIL + LOG_INFO, "Closed connection removed.\n");
						}
					else
						{
/* 0 terminate the input string */
						*(buf + t) = 0;
						session->process(buf);
						if (session->getstatus() == SMTP_STATUS_COMPLETE)
							{
							walk.del();
							}
/* Need to add code to handle received input */

						}
					}
				if (sock != -1 && FD_ISSET(sock, &fdw))
					{	
					if (session->dosend() < 0)
						{
						walk.del();
						syslog(LOG_MAIL + LOG_INFO, "Send failed. Connection removed.\n");
						}
					}
				} while(walk.next());
			}
		}
	}

if (walk.start())
	{
	do	{
		walk.del();
		syslog(LOG_MAIL + LOG_INFO, "Zombie link removed.");
		} while(walk.next());
	}

syslog(LOG_MAIL + LOG_NOTICE, PROGRAM " " VERSION " daemon shutdown");

return 0;
}


void help(void)
{
std::cout << COPYRIGHT << endl << endl;
std::cout << "Arguments:" << endl;
std::cout << "-h, --help            Displays this help" << endl;
std::cout << "-l, --list            List the outgoing queue" << endl;
std::cout << "-d, --daemon          Run as a mail server (daemon)" << endl;
std::cout << "-n, --nodaemon        Run as a mail server in the forground" << endl;
std::cout << "-q, --queue           Process and send outgoing mail" << endl;
std::cout << "-D<n>, --debug=<n>    Sets logging level (1 if no number specified)" << endl;
std::cout << "-r, --retry <t>       Set minimum time before resending a message" << endl;
std::cout << endl;
std::cout << "This program is released under the GNU General Public License (GPL) Version 2." << endl;
std::cout << "Please see the supplied file LICENSE, or the GNU Web-site at www.gnu.org" << endl;
}

int processitem(QUEUE *item)
{
char filename[PATH_MAX];
char line[SMTP_MAX_LINE];
TWALK<STRING> towalk;
int mature = 1;
int first = 1;
int fd;
int locked = 1;
char *dfile;
struct flock flk;
time_t now;

now = time(NULL);

if (!clientinitialise(domain->getsmtp()))
	return 0;

snprintf(filename, PATH_MAX-1, "%s/R%s", domain->getqout(), item->get());

if ((fd = open(filename, O_RDWR)) != -1)
	{
	fcntl(fd, F_GETLK, &flk);
	if (flk.l_type == F_UNLCK)
		locked = 0;
	close(fd);
	}

if (locked)
	{
	std::cout << "Locked" << endl;
	return 0;
	}

client->setfrom(item->getfrom());

towalk.attach(item->getto());
if (towalk.start())
	{
	do	{
		client->addto(towalk.get()->get());
		} while(towalk.next());
	}

towalk.attach(item->getdelivered());
if (towalk.start())
	{
	do	{
		client->adddelivered(towalk.get()->get());
		} while(towalk.next());
	}

client->setqtime(item->getqtime());

client->setbouncecount(item->getbouncecount());

if (item->getltime() != 0)
	{
	client->setltime(item->getltime());

	if ((now - item->getltime()) < retrytime)
		mature = 0;
	}

if (mature)
	{	
	client->openfiles(item->get());

	std::cout << "Sending... " << flush;
	
	if (clientrun(cached))
		std::cout << "Complete" << endl;
	else
		std::cout << "Failed" << endl;

	client->reset();

	}
else
	std::cout << filename << " Locked" << endl;
return 0;	
}

int loadqueue(char *dir, TLIST<QUEUE> *queue)
{
DIR *dlist;
int numitems = 0;

if ((dlist = opendir(dir)) == NULL)
	std::cout << "Could not open queue directory" << endl;
else
	{
	struct dirent *file;
	while((file = readdir(dlist)) != NULL)
		{
		if (*(file->d_name) == 'H')
			{
			numitems++;
			queue->add(NULL, new QUEUE(file->d_name+1));
			}
		}
	closedir(dlist);
	}
return numitems;
}

int queuerun(void)
{
char cwd[PATH_MAX];
char dir[PATH_MAX];
char line[SMTP_MAX_LINE];
TLIST<QUEUE> queue;
TWALK<QUEUE> walk;
int notprocessed = 0, numitems;

getcwd(cwd, PATH_MAX);

chdir(domain->getqout());
getcwd(dir, PATH_MAX);

numitems = loadqueue(dir, &queue);
queue.sort();

walk.attach(&queue);

std::cout << "Processing queue " << dir << endl;
std::cout << numitems << " items loaded for processing." << endl;

if (walk.start())
	{
	do	{
		std::cout << "Processing SMTP id " << walk.get()->get() << endl;
		notprocessed += processitem(walk.get());
		} while(walk.next());
	}
chdir(cwd);
if (client != NULL)
	client->sendquit();
return notprocessed;
}

void listqueue(void)
{
char cwd[PATH_MAX];
char dir[PATH_MAX];
char line[SMTP_MAX_LINE];
struct stat finfo;
TLIST<QUEUE> queue;
TWALK<QUEUE> walk;
char fname[PATH_MAX];
int numitems;
time_t now;

getcwd(cwd, PATH_MAX);

chdir(domain->getqout());
getcwd(dir, PATH_MAX);

numitems = loadqueue(dir, &queue);
queue.sort();

std::cout << "Listing queue " << dir << endl;
std::cout << numitems << " items loaded." << endl;

now = time(NULL);

walk.attach(&queue);
if (walk.start())
	{
	do	{
		TWALK<STRING> towalk;
		FILE *header;
		int first = 1;
		int fd;
		int locked = 1;
		struct flock flk;
		time_t time;

		snprintf(fname, PATH_MAX, "R%s", walk.get()->get());

		if ((fd = open(fname, O_RDWR)) != -1)
			{
			fcntl(fd, F_GETLK, &flk);
			if (flk.l_type == F_UNLCK)
				locked = 0;
			close(fd);
			}
		std::cout << "Message ID " << walk.get()->get() << ", size = " << walk.get()->getsize() << " (priority = " << walk.get()->priority() << ")" << endl;
				
		std::cout << (char *)(locked ? "* " : "  ") << setw(18) << "From" << " : " << walk.get()->getfrom() << endl;

		towalk.attach(walk.get()->getto());
		if (towalk.start())
			{
			do	{
				std::cout << setw(20) << (first ? "To" : "") << " : " << towalk.get()->get() << endl;
				first = 0;
				} while(towalk.next());
			}

		first = 1;
		towalk.attach(walk.get()->getdelivered());
		if (towalk.start())
			{
			do	{
				std::cout << setw(20) << (first ? "Complete" : "") << " : " << towalk.get()->get() << endl;
				first = 0;
				} while(towalk.next());
			}

		time = walk.get()->getqtime();
		std::cout << setw(20) << "Queue time" << " : " << ctime(&time);
		if ((time = walk.get()->getltime()) != 0)
			{
			std::cout << setw(20) << "Last attempt" << " : " << ctime(&time);
			std::cout << setw(20) << "Retry" << " : " << (char *)(((now-time)<retrytime) ? "Not expired" : "Ready to retry") << endl;
			}
		if (walk.get()->getresult() != 0 && *(walk.get()->getresult()) != 0)
			std::cout << setw(20) << "Last result" << " : " << walk.get()->getresult() << endl;

		if (walk.get()->getbouncecount() > 0)
			std::cout << setw(20) << "Bounce report" << " : " << walk.get()->getbouncecount() << endl;


		std::cout << endl;
		} while(walk.next());
	}
chdir(cwd);
}

int lookup(char *item, char *list[])
{
int n = 0;
while(list[n] != NULL)
	{
	if (strcmp(item, list[n]) == 0)
		return n;
	n++;
	}
return -1;
}

char *filename(char *path)
{
char *ptr;
if ((ptr = strrchr(path, '/')) == NULL)
	return path;
return ptr + 1;
}

//
// sendmail compatability mode (for queueing messages from stdin).
//
// Options:
// -F<from name>		Set the From full name
// -f<user>			Set the from user (root only)
// -i				Ignore dots on a line by themselves (normally terminates msg)
// -U				Initial (user) submission
// --				Trailing args are users

int sendmail(int argc, char *argv[])
{
int debug = 0, ignoredot = 0, users = 0, hasfrom = 0, usersubmit = 0, fullname = 0;
int n;
char *from = NULL;
char line[SMTP_MAX_LINE], tm[SMTP_MAX_LINE], ufullname[SMTP_MAX_LINE], ufrom[SMTP_MAX_LINE];
time_t t;
SMTPSERVER *client;
STRING_ARRAY tolist;

std::cout << "sendmail compatability mode" << endl;

for(n=1;n<argc;n++)
	{
	if (!users && *argv[n] == '-')
		{
		switch(*(argv[n]+1))
			{
			case 'b':
				if (*(argv[n]+2) == 'd')		// Don't run as sendmail demon
					return 0;
				break;
			case 'd':
				debug = strtol(argv[n]+2, NULL, 10);
				break;
			case 'i':
				ignoredot = 1;
				break;
			case 'U':
				usersubmit = 1;
				break;
			case '-':
				users = 1;
				break;
			case 'f':
				hasfrom = n;
				break;
			case 'F':
				fullname = n;
				break;
			}
		}
	else
		tolist.add(argv[n]);
	}

if ((client = new SMTPSERVER()) == NULL)
	return -1;

client->setdomain(domain);
client->serverstatus(220);
client->process("HELO VorpalMail\r\n");

client->setdebug(debug);

if (hasfrom && getuid() == 1)		// only root can be someone else
	strcpy(line, argv[hasfrom]+2);
else
	{
	struct passwd *pw;
	if ((pw = getpwuid(getuid())) == NULL)
		strcpy(line, "nobody");
	else
		strcpy(line, pw->pw_name);
	}
if (strchr(line, '@') == NULL)
	snprintf(ufrom, SMTP_MAX_LINE-1, "%s@%s", line, domain->getname());
else
	strcpy(ufrom, line);


if (fullname &&getuid() == 1)		// only root can be someone else
	strcpy(ufullname, argv[fullname]+2);
else
	{
	struct passwd *pw;
	if ((pw = getpwuid(getuid())) == NULL)
		strcpy(ufullname, "Unknown User");
	else
		strcpy(ufullname, pw->pw_gecos);
	if (*ufullname == 0)
		strcpy(ufullname, ufrom);
	}

snprintf(line, SMTP_MAX_LINE-1, "MAIL FROM:<%s>\r\n", ufrom);
client->process(line);

for(n=0;n<tolist.getsize();n++)
	{
	snprintf(line, SMTP_MAX_LINE-1, "RCPT TO:<%s>\r\n", tolist.get(n));
	client->process(line);
	}
client->process("DATA\r\n");

snprintf(line, SMTP_MAX_LINE-1, "From: \"%s\" <%s>\r\n", ufullname, ufrom);
client->process(line);


#ifdef VMAIL_GENERATE_TO
for(n=0;n<tolist.getsize();n++)
	{
	snprintf(line, SMTP_MAX_LINE-1, "%s %s%s\r\n", n==0 ? "To:" : "", tolist.get(n), n < (tolist.getsize()-1) ? "," : "");
	client->process(line);
	}
#endif

time(&t);
strftime(tm, SMTP_MAX_LINE-1, "%a, %d %b %Y %H:%M:%S %z", gmtime(&t));
snprintf(line, SMTP_MAX_LINE-1, "Date: %s\r\n", tm);
client->process(line);

while(fgets(line, SMTP_MAX_LINE-1, stdin) != NULL)
	{
	char *ptr;
	if ((ptr = strchr(line, '\n')) != NULL)
		*ptr = 0;
	if ((ptr = strchr(line, '\r')) != NULL)
		*ptr = 0;

	if (!ignoredot && strcmp(line, ".") == 0)
		break;
	
	if (*line == '.')
		snprintf(tm, SMTP_MAX_LINE-1, ".%s\r\n", line);
	else
		snprintf(tm, SMTP_MAX_LINE-1, "%s\r\n", line);

	client->process(tm);
	}
client->process(".\r\n");
client->process("QUIT\r\n");
return 0;
}

//
//
//
// Options:
//
// -l --list		show the queue contents
// -h --help		show a help message
// -d --daemon		run as a mail server daemon
// -q --queue		process the outgoing queue
//

int main(int argc, char *argv[])
{
int index = 0;
int val;
int mode = 0;

std::cout << PROGRAM << " " << VERSION << endl;

domain = new VIRTUAL_DOMAIN();
domain->load(CONFIG);

openlog(PROGRAM, LOG_CONS | LOG_PID, LOG_MAIL);

switch(lookup(filename(argv[0]), personality))
	{
	case PERS_MAILQ:
		listqueue();
		exit(0);
		break;
	case PERS_SENDMAIL:
	case PERS_SEND_MAIL:
		exit(sendmail(argc, argv));
		break;
	}

while((val = getopt_long(argc, argv, "r:lhdnqD::", longoptions, &index)) != -1)
	{
	switch(val)
		{
		case 'h':
			help();
			exit(1);
			break;
		case 'l':
			listqueue();
			exit(0);
			break;
		case 'd':
			if (mode == 0)
				mode = SMTP_SERVER;
			else
				mode = -1;
			break;
		case 'n':
			if (mode == 0)
				{
				nodaemon = 1;
				mode = SMTP_SERVER;
				}
			else
				mode = -1;
			break;
		case 'q':
			if (mode == 0)
				mode = SMTP_CLIENT;
			else
				mode = -1;
			break;
		case 'D':
			if (optarg != NULL)
				debug = strtol(optarg, NULL, 10);
			else
				debug = 1;
			break;
		case 'm':
			domain->dump();
			exit(2);
		case 'r':
			{
			long n;
			if ((n = strtol(optarg, NULL, 10)) < 1 || n > 4320)
				std::cout << "retry time must be in the range 1 - 4320 (minutes)" << endl;
			else
				retrytime = 60*n;
			}
			break;
		}
	index++;
	}

if (debug)
	std::cout << "Debug mode: " << debug << endl;

switch(mode)
	{
	case SMTP_SERVER:
		exit(serverrun());
	case SMTP_CLIENT:
		exit(queuerun());
	}

std::cout << "Must specify either --daemon or --queue mode." << endl;
std::cout << argv[0] << " --help for arguments." << endl;

closelog();
}

