
//
//
// Project       : Vorpal-Mail
//
// Source        : smtp.cc
//
// Version       : 1.0
//
// Creator       : Blane Bramble                    Date          : 23rd October 2000
// Amended       : Blane Bramble                    Date          : 27th March 2001
//
// License       : Copyright Virtua-Web 2000-2001, licensed under GNU General Public License (GPL) v2
//
// Description   :
//
// Class implementation for SMTP session management. Provides a single SMTP session capable
// of transferring email compliant with RFC 821
//
// Changelog     :
//
// 9th Nov 2000
// Record locking code added - all files should be locked exclusive whilst in use (using fcntl).
// Session naming code altered.
//
// 13th Nov 2000
// Preliminary code added to fork() before delivering messages. If the fork fails then the
// message will be delivered in the main process.
//
// 22nd Nov 2000
// Added code to correctly cleanup received queue once a message has been delivered. Also
// initial code to call an external virus scanner.
//
// 5th December 2000
// Additional code to parse headers and automatically revert to body text if a header is not
// in the form "header: text" (where header is a single word). Needed by the command line
// sendmail compatability
//
// 7th December 2000
// Added support to reset (RSET) the connection at the start of each message,
// as if a RCPT TO failed, the next message was being injected into the
// aborted session.
//
// 20th December 2000
// Added code to allow headers to be loaded into LIST::header structure.
//
// 16th March 2001
// Renamed smtp.cc from session.cc

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <pwd.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <syslog.h>
#include <errno.h>
#include <fcntl.h>

#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/wait.h>


#include "mail.h"
#include "array.h"
#include "smtp.h"
#include "virtual.h"

//
// Static initialisation:
//

char *SMTP_SESSION::commands[] = {
		"QUIT", 
		"HELP",
		"HELO",
		"EHLO",
		"MAIL",
		"RCPT",
		"DATA",
		"RSET",
		"NOOP",
		"VRFY",
		"EXPN",
		NULL };

char *SMTP_SESSION::command_help[] = {
	"Exit from " PROGRAM,
	"Display this help",
	"Request connection",
	"Request connection",
	"Start email from user",
	"Add user to recipient list",
	"Start transferring email data",
	"Reset session",
	"Do nothing",
	"Verify an email address",
	"Expand an email address",
	};

char *SMTP_SESSION::command_arg[] = {
	"",
	"",
	"<hostname>",
	"<hostname>",
	"FROM:<user@host>",
	"TO:<user@host>",
	"<data>",
	"",
	"",
	"<user@host>",
	"<user@host>",
	};


char *SMTP_SESSION::hostname = NULL;

VIRTUAL_DOMAIN *SMTP_SESSION::domain = NULL;

long SMTP_SESSION::sequence = -1;
long SMTP_SESSION::pid = -1;
int SMTP_SESSION::debug = 0;

//
// Private class functions:
//

void SMTP_SESSION::initialise(void)
{

if (SMTP_SESSION::hostname == NULL)
	{
	char buf[128];
	gethostname(buf, 127);
	SMTP_SESSION::hostname = strdup(buf);
	}

if (SMTP_SESSION::sequence == -1)
	SMTP_SESSION::sequence = 1;

if (SMTP_SESSION::pid == -1)
	SMTP_SESSION::pid = getpid();

bouncecount = bounce = bouncing = 0;

bodyfile = NULL;
bouncefile = NULL;
envelopefile = NULL;

sendcount = 0;

filename = "";
line = NULL;
prev = next = NULL;
socket = -1;
*spill = 0;
status = SMTP_STATUS_OPEN;		// waiting for HELO

from = "";

dataheader = 1;
datamime = 0;
datamimesection = NULL;

ltime = qtime = 0;

lastheader = NULL;
}

void SMTP_SESSION::addbounce(char *fmt, ...)
{
va_list argptr;

if (bouncing)
	return;

va_start(argptr, fmt);
vfprintf(bouncefile, fmt, argptr);
bounce++;
va_end(argptr);
}


void SMTP_SESSION::closefiles(void)
{
if (bodyfile != NULL)
	{
	unlock(bodyfile);
	fclose(bodyfile);
	}
bodyfile = NULL;

if (bouncefile != NULL)
	{
	unlock(bouncefile);
	fclose(bouncefile);
	}
bouncefile = NULL;

if (envelopefile != NULL)
	{
	unlock(envelopefile);
	fclose(envelopefile);
	}
envelopefile = NULL;
}

void SMTP_SESSION::addheader(char *key, char *fmt, ...)
{
va_list argptr;
char scratch[1024];
FILE *header;			// needs to add directly to envelop file, cos headers have been written to file at this point

//vsnprintf(scratch, 1023, fmt, argptr);
//header->add(key, scratch);

va_start(argptr, fmt);
if ((header = fopen(makefilename(domain->getqin(), 'E', filename.get()), "a")) != NULL)
	{
	fprintf(header, "%s: ", key);
	vfprintf(header, fmt, argptr);
	fclose(header);
	}
va_end(argptr);
}

void SMTP_SESSION::reset(void)
{
int i;
char *ptr;
TLISTITEM<STRING> *item;

bouncecount = bounce = bouncing = 0;

*spill = 0;
dataheader = 1;
datamime = 0;

closefiles();

filename = "";

line = NULL;

status = SMTP_STATUS_HELO;

from = "";

if (datamimesection != NULL)
	free(datamimesection);

while((item = header.gethead()) != NULL)
	header.del(item);

while(to.getsize() > 0)
	to.remove(0);
while(delivered.getsize() > 0)
	delivered.remove(0);
}

void SMTP_SESSION::closesocket(void)
{
reset();
if (socket != -1)
	{
	shutdown(socket, 2);
	close(socket);
	}
socket = -1;
}

int SMTP_SESSION::lookup(char *s)
{
int n=0;
while(SMTP_SESSION::commands[n] != NULL)
	{
	if (strcasecmp(SMTP_SESSION::commands[n], s) == 0)
		return n;
	n++;
	}
return -1;
}

int SMTP_SESSION::splitheader(char *line, char *h, char *t)
{
char *sfrom, *sto;
if (strchr(line, ':') == NULL)
	return 0;

sfrom = line;
sto = h;

while(isspace(*sfrom))
	sfrom++;

while(*sfrom != 0 && *sfrom != ':')
	*sto++ = *sfrom++;
*sto = 0;

if (*sfrom == ':')
	sfrom++;
while(isspace(*sfrom))
	sfrom++;

sto = t;
while(*sfrom != 0)
	*sto++ = *sfrom++;
*sto = 0;
if (*t != 0)
	return 2;
return 1;
}

int SMTP_SESSION::parsename(char *sfrom, char *pre, char *sto)
{
while(*pre != 0)		// check the correct prefix is present
	{
	if (toupper(*sfrom) != toupper(*pre))
		return 0;
	sfrom++; pre++;
	}

while(isspace(*sfrom) && *sfrom != 0)
	sfrom++;

if (*sfrom++ != '<')		// not in angle brackets
	return 0;

while(*sfrom != '>')
	*sto++ = *sfrom++;
*sto = 0;
return 1;
}

int SMTP_SESSION::localuser(char *msgto, VIRTUAL_DOMAIN *dom)
{
int statuslocal = 1, statusrelay = -1, statusinvalid = 0;		// possible results for local user, remote user, local but doesnt exist

if (dom != NULL)
	{
	switch(dom->getunknown())
		{
		case VIRTDOM_UNKNOWN_ACCEPT:							// invalid treated as local
			statusinvalid = 1;
			break;
		case VIRTDOM_UNKNOWN_REJECT:							// default settings
			break;
		case VIRTDOM_UNKNOWN_RELAY:							// invalid treated as remote
			statusinvalid = -1;
			break;
		case VIRTDOM_UNKNOWN_MAILTO:							// invalid treates as alias
			if (getpwnam(dom->getunknownrelay()) != NULL)				// mailto is a local user
				statusinvalid = 2;						// so replace any unknown users
			else
				statusinvalid = 0;						// otherwise reject unknown
			break;
		}
	if (getpwnam(msgto) == NULL)
		return statusinvalid;
	else
		return statuslocal;
	}
else
	return statusrelay;										// remote user
}


int SMTP_SESSION::lock(FILE *fp)
{
struct flock flk;

flk.l_type = F_WRLCK;
flk.l_whence = 0;
flk.l_start = 0;
flk.l_len = 0;

if (fcntl(fileno(fp), F_SETLK, &flk) == -1)
	return 0;
return 1;
}

void SMTP_SESSION::unlock(FILE *fp)
{
struct flock flk;

flk.l_type = F_UNLCK;
flk.l_whence = 0;
flk.l_start = 0;
flk.l_len = 0;

fcntl(fileno(fp), F_SETLK, &flk);
}


void SMTP_SESSION::addreceive(void)
{
long l;
int n, quad[4];
time_t t;
char tm[128];
char scratch[1024];
char *hname ="unknown";
struct hostent *host;

time(&t);
strftime(tm, sizeof(tm), "%a, %d %b %Y %H:%M:%S %z", gmtime(&t));

l = ntohl(saddr.sin_addr.s_addr);
for(n=0;n<4;n++)
	{
	quad[n] = l & 0xff;
	l >>= 8;
	}

if ((host = gethostbyaddr((char *)&saddr.sin_addr.s_addr, sizeof(saddr.sin_addr.s_addr), AF_INET)) != NULL)
	hname = host->h_name;

snprintf(scratch, 1023, "from %s (%s [%d.%d.%d.%d]) by %s (" PROGRAM " " VERSION ") with SMTP id %s; %s", hname, hname, quad[3], quad[2], quad[1],
		quad[0], SMTP_SESSION::hostname, filename.get(), tm);
header.add("Received", new STRING(scratch));
}

char *SMTP_SESSION::sessionname(void)
{
static char namebuf[PATH_MAX];
time_t ut;

time(&ut);

// session name is unixtime shifted right by 12 bits and then the low 16 bits masked (which correspond to bits 12-28 of the original)
// this main section will thus change every 68 minutes, wrapping every 3106 days (by which time any stuck message should have bounced, so
// the time:sequence pairing should never be duplicated).

snprintf(namebuf, PATH_MAX-1, "%.4x%.4x%.6x", SMTP_SESSION::pid, (unsigned int)((ut>>12) & 0xffff), (int)SMTP_SESSION::sequence++);

return namebuf;
}


//
// PUBLIC class functions:
//

SMTP_SESSION::SMTP_SESSION(void)
{
initialise();
// Need to make this point to the local machine by default
saddr.sin_family = AF_INET;
saddr.sin_addr.s_addr = htonl(0x7f000001);
saddr.sin_port = htons(0);
}

SMTP_SESSION::SMTP_SESSION(int s, struct sockaddr_in *sa)
{
initialise();
memcpy(&this->saddr, sa, sizeof(struct sockaddr_in));
socket = s;
}

SMTP_SESSION::~SMTP_SESSION()
{
closesocket();			// does reset etc.
}

void SMTP_SESSION::setnext(SMTP_SESSION *n)
{
this->next = n;
n->prev = this;
}

//
// SMTP_SESSION forms a file name by combining a queue directory, a file type and the session name.
//
// file types are:
//
// 'M'		Message body received in queuein
// 'R'		Message body relayed to queueout
// 'H'		SMTP_SESSION header information
// 'E'		RFC 822 header information
// 'F'		RFC 822 header relayed to queueout
// 'B'		Bounce information
//

char *SMTP_SESSION::makefilename(char *queue, char type, char *file)
{
static char path[PATH_MAX];

snprintf(path, PATH_MAX-1, "%s/%c%s", queue, type, file);

return path;
}



void SMTP_SESSION::setfrom(char *f)
{
from = f;
}

void SMTP_SESSION::addto(char *t)
{
if (to.find(t) == -1)
	to.add(t);
}


int SMTP_SESSION::sendf(char *fmt, ...)
{
char tmp[1024];
va_list argptr;
va_start(argptr, fmt);
vsnprintf(tmp, 1023, fmt, argptr);

sendbuffer.add(NULL, new STRING(tmp));

if ((SMTP_SESSION::debug > 2) || (SMTP_SESSION::debug > 1 && this->status != SMTP_STATUS_DATASEND))
	syslog(LOG_MAIL + LOG_DEBUG, ">> %s\n", tmp);

va_end(argptr);

return 1;
}

int SMTP_SESSION::ispending(void)
{
if (sendbuffer.gethead() == 0)
	return 0;
return 1;
}

int SMTP_SESSION::dosend(void)
{
if (ispending())
	{
	TLISTITEM<STRING> *item;
	if ((item = sendbuffer.gethead()) != 0)
		{
		char *ptr;
		int tosend, count;

		ptr = item->getdata()->get();

		tosend = strlen(ptr) - sendcount;

		if ((count = send(socket, ptr + sendcount, tosend, 0)) < 0)
			return count;
		else if ((sendcount += count) == tosend)
			{
			sendbuffer.del(item);
			sendcount = 0;
			}
		return count;
		}
	}
return 0;
}

void SMTP_SESSION::writeheader(void)
{
writeheader(-1);
}

void SMTP_SESSION::writeheader(int max)
{
FILE *hfp;
int n = 0;

if (to.getsize() == 0)
	{
	if (debug > 1)
		syslog(LOG_MAIL + LOG_INFO, "** No remote users to write header for.\n");
	return;
	}

if ((hfp = fopen(makefilename(domain->getqout(), 'H', filename.get()), "w")) == NULL)
	{
	syslog(LOG_MAIL + LOG_DEBUG, "Error opening header to write");
	return;
	}

lock(hfp);
fprintf(hfp, "F%s\n", from.get());

while(to.get(n) != 0 && (max == -1 || n < max))
	{
	fprintf(hfp, "T%s\n", to.get(n));
	n++;
	}

n = 0;
while(delivered.get(n) != NULL)
	{
	fprintf(hfp, "D%s\n", delivered.get(n));
	n++;
	}

if (this->qtime != 0)
	fprintf(hfp, "Q%ld\n", qtime);

if (this->ltime != 0)
	fprintf(hfp, "L%ld\n", ltime);
if (last_response.get() != 0)
	fprintf(hfp, "R%s\n", last_response.get());
if (bouncecount)
	fprintf(hfp, "B%ld\n", bouncecount);

unlock(hfp);
fclose(hfp);
}


void SMTP_SESSION::setdomain(VIRTUAL_DOMAIN *d)
{
domain = d;
}

void SMTP_SESSION::adddelivered(char *who)
{
delivered.add(who);
}

