
//
//
// Project       : Vorpal-Mail
//
// Source        : smtp.h
//
// Version       : 1.0
//
// Creator       : Blane Bramble                    Date          : 23rd October 2000
// Amended       : Blane Bramble                    Date          : 27th March 2001
//
// License       : Copyright Virtua-Web 2000-2001, licensed under GNU General Public License (GPL) v2
//
// Description   :
//
// Definition files for SMTP session class. Supports a single SMTP session, parsing
// the relevant commands to provide an RFC 821 compliant session.
//
// Changelog     :
//
// 20th Dec 2000
// Added class members to allow headers to be loaded (and parsed). Changed a couple of char ptr's to arrays as they
// were malloc'd for the life of the class instance anyway.
//
// 16th March 2001
// Renamed smtp.h from session.h

#include "smtpdef.h"
#include "virtual.h"
#include "string.h"

#ifndef __SMTP_SESSION__
#define __SMTP_SESSION__

enum SMTP_CODE {
	SMTP_ERROR = -1,
	SMTP_OK = 0,
	};

enum SMTP_STATUS {
	SMTP_STATUS_OPEN = 0,

	SMTP_STATUS_HELO,		// awaiting HELO
	SMTP_STATUS_MAIL,		// got HELO, awaiting MAIL (start of msg)
	SMTP_STATUS_RCPT,		// got MAIL, awaiting RCPT or DATA
	SMTP_STATUS_DATA,		// in DATA, waiting for .
	SMTP_STATUS_RSET,

	SMTP_STATUS_DATATEXT,
	SMTP_STATUS_SENT,		// client only, msg sent
	SMTP_STATUS_DATASEND,		// client only, still sending data
	SMTP_STATUS_SENDRETRY,		// client only, didnt buffer, so wait and retry

	SMTP_STATUS_COMPLETE = 100,

	SMTP_STATUS_TIMEOUT = -2,	// used to process timeouts

	SMTP_STATUS_ABORT = -1,
	

	};

enum {
	SMTP_CMD_QUIT = 0,
	SMTP_CMD_HELP,
	SMTP_CMD_HELO,
	SMTP_CMD_EHLO,
	SMTP_CMD_MAIL,
	SMTP_CMD_RCPT,
	SMTP_CMD_DATA,
	SMTP_CMD_RSET,
	SMTP_CMD_NOOP,
	SMTP_CMD_VRFY,
	SMTP_CMD_EXPN,
	};

class SMTP_SESSION {
	protected:
		static char *commands[];
		static char *command_help[];
		static char *command_arg[];
		static char *hostname;
		static long sequence;
		static long pid;
		static int debug;
		static VIRTUAL_DOMAIN *domain;

		STRING last_response;
		int bounce, bouncing;

		int bouncecount;

		STRING_ARRAY delivered;

		STRING filename;
		FILE *bodyfile, *bouncefile, *envelopefile;
		int socket;
		SMTP_STATUS status;
		char *line;
		char scratch[SMTP_MAX_LINE*2];
		char spill[SMTP_MAX_LINE*2];
		SMTP_SESSION *next, *prev;
		struct sockaddr_in saddr;

		TLIST<STRING> sendbuffer;
		int sendcount;
	
		TLIST<STRING> header;
		TLISTITEM<STRING> *lastheader;

		STRING from;
		STRING_ARRAY to;

		int loop;
		int dataheader, datamime;
		char *datamimesection;

		long qtime, ltime;

		int timeout;

		void initialise(void);
		int sendf(char *, ...);
		int lookup(char *);
		void process_cmd(char *);
		int parsename(char *, char *, char *);
		void addreceive(void);
		char *sessionname(void);
		char *makefilename(char *, char, char *);
		int localuser(char *, VIRTUAL_DOMAIN *);
		void closefiles(void);
		int lock(FILE *);
		void unlock(FILE *);
		void addheader(char *, char *, ...);
		int splitheader(char *, char *, char *);
		void addbounce(char *, ...);
		void writeheader(void);
		void writeheader(int);
	public:
		SMTP_SESSION(void);
		SMTP_SESSION(int, struct sockaddr_in *);
		~SMTP_SESSION();
		char *getfilename(void) { return filename.get(); };
		SMTP_STATUS getstatus(void) { return status; };
		int getsocket(void) { return socket; };
		void setsocket(int s) { socket = s; };
		void setnext(SMTP_SESSION *);
		SMTP_SESSION *getnext(void) { return next; };
		SMTP_SESSION *getprev(void) { return prev; };
		void closesocket(void);
		void setfrom(char *);
		void addto(char *);
		void setdebug(int n) { debug = n; };
		void setdomain(VIRTUAL_DOMAIN *);
		void reset(void);
		int ispending(void);
		int dosend(void);
		void setqtime(long t) { qtime = t; };
		void setltime(long t) { ltime = t; };
		void setbouncecount(int n) { bouncecount = n; };
		void adddelivered(char *);
	};


#endif

